using System;
using System.Linq;
using Neo4RestNet.Core;

namespace Neo4RestNet.ObjectMapping
{
  public class ObjectMap : IObjectMap
  {
    private readonly IReflector _reflector;
    private readonly IPropertyMap _propertyMap;
    private readonly IGraphDatabaseService _graphDatabaseService;
    // Property stored on nodes to be able to recreate the correct class
    private readonly string _concreteClassPropertyName;
    // Property on objects to be able to map the object back to its node
    private readonly string _nodeReferencePropertyName;

    public ObjectMap(IReflector reflector, IPropertyMap propertyMap, IGraphDatabaseService graphDatabaseService, string concreteClassPropertyName, string nodeReferencePropertyName)
    {
      _reflector = reflector;
      _propertyMap = propertyMap;
      _graphDatabaseService = graphDatabaseService;
      _concreteClassPropertyName = concreteClassPropertyName;
      _nodeReferencePropertyName = nodeReferencePropertyName;
    }

    private Property GetNodeReferenceProperty(object o)
    {
      var selfProperty = _reflector.GetProperties(o.GetType()).Where(p => p.Name == _nodeReferencePropertyName).FirstOrDefault();
      if (selfProperty == null) throw new Exception(string.Format("Class {0} does not implement property {1} that is required to map object to node", o.GetType().Name, _nodeReferencePropertyName));
      return selfProperty;
    }

    private string GetConcreteClass(INode node)
    {
      // Retrieve the concrete class that has been saved to the node
      var concreteClass = node.GetProperty<string>(_concreteClassPropertyName, null);
      if (concreteClass == null) throw new Exception(string.Format("Node {0} is missing attribute {1} that is required to map node to object", node.Self, _concreteClassPropertyName));
      return concreteClass;
    }

    private INodeReference CreateNodeReference(string self)
    {
      return new RestNodeReference(self);
    }

    public INode GetNodeFromObject(object from)
    {
      // Get the objects nodereference that in turn will be able to get the node
      return ((INodeReference)GetNodeReferenceProperty(from).Get(from)).GetNode();
    }

    public object GetObjectFromNode(INode from)
    {
      // Retrieve the concrete class that has been saved to the node
      var concreteClass = GetConcreteClass(from);
      // Invoke empty constructor to get a new object
      var theObject = _reflector.GetEmptyConstructor(concreteClass).Invoke();
      // Map all properties on node to the object
      _propertyMap.Container2Object(from, theObject);
      // Let the instance know how to reference its own node
      GetNodeReferenceProperty(theObject).Set(theObject, CreateNodeReference(from.Self));

      return theObject;
    }

    public INode CopyObjectToNode(object from)
    {
      // Get the node from the object through its node reference property
      var node = ((INodeReference)GetNodeReferenceProperty(from).Get(from)).GetNode();
      // Save the concrete class of the object
      node.SetProperty(_concreteClassPropertyName, from.GetType().Name);
      // Get properties on object & set them on the node
      _propertyMap.ObjectToContainer(from, node);

      return node;
    }

    public INode CreateNodeFromObject(object newObject)
    {
      var properties = new PropertyContainer();
      // Save the concrete class of the object
      properties.SetProperty(_concreteClassPropertyName, newObject.GetType().Name);
      // Get properties on object & set them on the property container
      _propertyMap.ObjectToContainer(newObject, properties);
      // Save the new node
      var node = _graphDatabaseService.CreateNode(properties);
      // Let the instance know how to retrieve its own node
      GetNodeReferenceProperty(newObject).Set(newObject, CreateNodeReference(node.Self));

      return node;
    }
  }
}