using System.Collections.Generic;

namespace Neo4RestNet.Core
{
  public class RelationshipCondition
  {
    public string RelationshipName { get; set; }
    public Direction Direction { get; set; }
  }

  public enum TraversalOrder
  {
    BreadthFirst = 0,
    DepthFirst = 1,
  }

  public enum StopEvaluator
  {
    DepthOne,
    EndOfGraph,
  }

  public enum ReturnableEvaluator
  {
    All = 0,
    AllButStartNode = 1,
  }

  public enum Direction
  {
    Incoming = 0,
    Outgoing = 1,
    All = 2,
  }

  public interface INode : IPropertyContainer
  {
    /// <summary>
    /// Deletes the current node
    /// </summary>
    void Delete();

    /// <summary>
    /// Returns all outgoing and ingoing relationships attached to the current node
    /// </summary>
    /// <returns></returns>
    IEnumerable<IRelationship> GetRelationships();

    /// <summary>
    /// Returns the requested relations from the current node
    /// </summary>
    /// <param name="direction">Direction indicates if the relations should be from (outgoing) or to (incoming) the current node</param>
    /// <returns></returns>
    IEnumerable<IRelationship> GetRelationships(Direction direction);

    /// <summary>
    /// Returns all outgoing and incoming relations with the specified name from the current node
    /// </summary>
    /// <param name="names"></param>
    /// <returns></returns>
    IEnumerable<IRelationship> GetRelationships(IEnumerable<string> names);

    /// <summary>
    /// Returns the relationship, null if none found or throws if more than one found
    /// </summary>
    /// <param name="name"></param>
    /// <param name="direction"></param>
    /// <returns>Null if no relationship was found</returns>
    IRelationship GetSingleRelationship(string name, Direction direction);

    bool HasRelationship();
    bool HasRelationship(Direction direction);
    bool HasRelationship(IEnumerable<string> names);
    bool HasRelationship(string name, Direction direction);

    /// <summary>
    /// Adds a new relation from the current node to the specified node. (Outgoing direction)
    /// </summary>
    /// <param name="to">The destination node</param>
    /// <param name="name">Name of the relation</param>
    IRelationship CreateRelationshipTo(INode to, string name);

    /// <summary>
    /// Returns a list of nodes as a result of the traverser algorithm specified
    /// </summary>
    /// <param name="traverser">Traverser algorithm as a string</param>
    /// <returns>Enumeration of nodes</returns>
    IEnumerable<INode> TraverseNodes(string traverser);

    IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, StopEvaluator stopEvaluator,
                                     ReturnableEvaluator returnableEvaluator, string relationshipName, Direction direction);
    IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, string stopEvaluator,
                                     ReturnableEvaluator returnableEvaluator, string relationshipName, Direction direction);
    IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, string stopEvaluator,
                                     string returnableEvaluator, string relationshipName, Direction direction);
    IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, StopEvaluator stopEvaluator,
                                     string returnableEvaluator, string relationshipName, Direction direction);

    /// <summary>
    /// Name in native java interface is getId()
    /// </summary>
    long Id { get; }

    /// <summary>
    /// Full path to current node
    /// This is not part of native java interface
    /// </summary>
    string Self { get; }
  }
}