﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using Neo4RestNet.Core;

namespace Neo4RestNet.ExpressionParser
{
  public interface ITraversalPosition
  {
    INode Node { get; }   // Another node type here to better match what can be done?
    int Depth { get; }
    bool IsStartNode { get; }
    IRelationship LastRelationshipTraversed { get; } // Another relationship type here to better match what can be done?
    bool NotStartNode { get; }
    INode PreviousNode { get; }   // Another node type here to better match what can be done?
    int ReturnedNodesCount { get; }
  }

  public class ExpressionParser
  {
    private readonly AtomConverter _atomConverter;
    private readonly ExpressionEvaluator _expressionEvaluator;

    private readonly IDictionary<string, IReferenceResolver> _referenceResolvers =
      new Dictionary<string, IReferenceResolver>
        {
          {typeof (ITraversalPosition).ToString(), new JsNetConventionReferenceResolver()},
          {typeof (INode).ToString(), new JsNetConventionReferenceResolver()},
          {typeof(string).ToString(), new StringReferenceResolver()}
        };

    private readonly IParameterResolver _parameterResolver = new NeoEvaluatorParameterResolver();

    public ExpressionParser(AtomConverter atomConverter, ExpressionEvaluator expressionEvaluator)
    {
      _atomConverter = atomConverter;
      _expressionEvaluator = expressionEvaluator;
    }

    public string ConvertToJs(Expression<Func<ITraversalPosition, bool>> lambdaExpression)
    {
      return Parse(lambdaExpression.Body);
    }

    private string ParseConstant(ConstantExpression constantExpression)
    {
      // Converts to atom or throws!
      return _atomConverter.Convert(constantExpression.Value);
    }

    private string ParseBinary(BinaryExpression expression)
    {
      return Parse(expression.Left) + GetBinaryOperation(expression) + Parse(expression.Right);
    }

    private string ParseCall(MethodCallExpression methodCallExpression)
    {
      var receivingObjectName = methodCallExpression.Object != null ? methodCallExpression.Object.Type.ToString() : "static";

      if (_referenceResolvers.Keys.Contains(receivingObjectName))
      {
        return BuildCall(methodCallExpression, _referenceResolvers[receivingObjectName]);
      }

      // Try evaluating the method call
      return _expressionEvaluator.EvaluateMethodCall(methodCallExpression);
    }


    private string BuildCall(MethodCallExpression methodCallExpression, IReferenceResolver referenceResolver)
    {
      var builder = new StringBuilder("");

      builder.Append(Parse(methodCallExpression.Object));
      builder.Append(".");
      builder.Append(referenceResolver.GetMethodName(methodCallExpression));
      builder.Append("(");
      builder.Append(string.Join(",",
        methodCallExpression.Arguments.Select(a => Parse(a)).ToArray()));
      builder.Append(")");

      return builder.ToString();
    }

    private static string GetBinaryOperation(Expression binaryOperation)
    {
      switch (binaryOperation.NodeType)
      {
        case ExpressionType.Equal:
          return " == ";
        case ExpressionType.NotEqual:
          return " != ";
        case ExpressionType.GreaterThan:
          return " > ";
        case ExpressionType.GreaterThanOrEqual:
          return " >= ";
        case ExpressionType.LessThan:
          return " < ";
        case ExpressionType.LessThanOrEqual:
          return " <= ";
        case ExpressionType.AndAlso:
          return " && ";
        case ExpressionType.OrElse:
          return " || ";
        case ExpressionType.And:
          return "&";
        case ExpressionType.Or:
          return "|";
        case ExpressionType.Add:
        case ExpressionType.AddChecked:
          return " + ";
        case ExpressionType.Subtract: 
        case ExpressionType.SubtractChecked:
          return " - ";
      }
      throw new Exception();
    }

    private string ParseMemberAccess(MemberExpression memberExpression)
    {
      var innerExpression = memberExpression.Expression;

      switch (innerExpression.NodeType)
      {
        case ExpressionType.Parameter:
          // Accessing a member on an object passed in as a parameter to the expression
          return _parameterResolver.GetParameterName((ParameterExpression)innerExpression) +
                 "." + _referenceResolvers[innerExpression.Type.ToString()].GetMemberName(memberExpression);

        case ExpressionType.Constant:
          // Accessing a member on an object in outer scope
          return _expressionEvaluator.EvaluateMemberAccess(memberExpression);

        default:
          throw new Exception("Unknown type of memberaccess");
      }
    }

    public string Parse(Expression expression)
    {
      switch (expression.NodeType)
      {
        case ExpressionType.Call:
          return ParseCall((MethodCallExpression) expression);
        
        case ExpressionType.Constant:
          return ParseConstant((ConstantExpression) expression);

        case ExpressionType.MemberAccess:
          return ParseMemberAccess((MemberExpression)expression);

        case ExpressionType.LessThan:
        case ExpressionType.LessThanOrEqual:
        case ExpressionType.GreaterThan:
        case ExpressionType.GreaterThanOrEqual:
        case ExpressionType.NotEqual:
        case ExpressionType.Equal:
        case ExpressionType.AndAlso:
        case ExpressionType.Add:
        case ExpressionType.AddChecked:
        case ExpressionType.Subtract:
        case ExpressionType.SubtractChecked:
        case ExpressionType.OrElse:
          return ParseBinary((BinaryExpression)expression);

        default:
          throw new ArgumentOutOfRangeException();
      }
    }
  }
}