using System.Collections.Generic;
using System.Linq;
using System.Net;
using Neo4RestNet.Core;
using Neo4RestNet.Rest.Encoders;
using Neo4RestNet.Rest.Parsers;
using Neo4RestNet.Rest.Traverser;

namespace Neo4RestNet.Rest
{
  public class RestNode : INode
  {
    private PropertyContainer _propertyContainer;

    private static readonly IDictionary<Direction, string> _directionNames
      = new Dictionary<Direction, string>
          {
            { Direction.Incoming, "in" },
            { Direction.Outgoing, "out" },
            { Direction.All, "all" },
          };

    public RestNode(string databaseUrl, long id)
    {
      Id = id;
      Self = databaseUrl + "/node/" + id;
    }

    public RestNode(string self)
    {
      InitializeFromSelf(self);
    }

    public RestNode(string self, string jsonData)
    {
      InitializeFromSelf(self);
      _propertyContainer = CreatePropertyContainer(jsonData);
    }

    internal RestNode(string self, IDictionary<string, object> properties)
    {
      InitializeFromSelf(self);
      _propertyContainer = new PropertyContainer(properties);
    }

    private void InitializeFromSelf(string self)
    {
      // Make sure there is no trailing /
      Self = self.TrimEnd('/');
      // Parse out id, should be last part
      Id = long.Parse(self.Split('/').Last());      
    }

    private static RestServer Server { get { return RestServer.Instance(); } } 

    public void Delete()
    {
      try
      {
        Server.Delete(Self);
      }
      catch (WebException e)
      {        
        throw new GraphDbException(string.Format("Failed to delete node {0}", Id), e);
      }
    }

    public IEnumerable<IRelationship> GetRelationships()
    {
      return GetRelationships(Direction.All);
    }

    public IEnumerable<IRelationship> GetRelationships(Direction direction)
    {
      var response = Server.Get(Self + "/relationships/" + _directionNames[direction]);
      return new RelationshipParser().ParseArray(response);
    }

    public IEnumerable<IRelationship> GetRelationships(IEnumerable<string> names)
    {
      var response = Server.Get(Self + "/relationships/" + _directionNames[Direction.All] + "/" + string.Join("&", names.ToArray())  );
      return new RelationshipParser().ParseArray(response);
    }

    public IRelationship GetSingleRelationship(string name, Direction direction)
    {
      var response = Server.Get(Self + "/relationships/" + _directionNames[direction] + "/" + name);
      var relationships = new RelationshipParser().ParseArray(response);

      if (relationships.Count() > 1)
        throw new GraphDbException(
          string.Format("There were more than one relationship on node {0} when a single relationship was requested", Self));

      return relationships.FirstOrDefault();
    }

    public bool HasRelationship()
    {
      // Not the most efficient but good enough right now
      return GetRelationships().Count() > 0;
    }

    public bool HasRelationship(Direction direction)
    {
      // Not the most efficient but good enough right now
      return GetRelationships(direction).Count() > 0;
    }

    public bool HasRelationship(IEnumerable<string> names)
    {
      // Not the most efficient but good enough right now
      return GetRelationships(names).Count() > 0;
    }

    public bool HasRelationship(string name, Direction direction)
    {
      // Not the most efficient but good enough right now
      return GetRelationships(direction).Count(r => r.Name == name) > 0;
    }

    public IRelationship CreateRelationshipTo(INode to, string name)
    {
      // Hmm.. strange warning when using String.Format(..)
      var body = @"{""to"":""{0}"", ""type"":""{1}""}".Replace("{0}", to.Self).Replace("{1}", name);
      var response = Server.Post(Self + "/relationships", body);
      return new RelationshipParser().Parse(response);
    }

    public IEnumerable<INode> TraverseNodes(string traverser)
    {
      var response = Server.Post(Self + "/traverse/node", traverser);
      return new NodeArrayParser().Parse(response);
    }

    public IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, StopEvaluator stopEvaluator, ReturnableEvaluator returnableEvaluator, string relationshipName, Direction direction)
    {
      var traversalBuilder = new TraversalBuilder();
      traversalBuilder.SetOrder(traversalOrder);
      traversalBuilder.SetStopEvaluator(stopEvaluator);
      traversalBuilder.SetReturnableEvaluator(returnableEvaluator);
      traversalBuilder.AddRelationshipCondition(new RelationshipCondition {Direction = direction, RelationshipName = relationshipName});
      var traverser = traversalBuilder.Build();
      return TraverseNodes(traverser);
    }

    public IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, string stopEvaluator, ReturnableEvaluator returnableEvaluator, string relationshipName, Direction direction)
    {
      var traversalBuilder = new TraversalBuilder();
      traversalBuilder.SetOrder(traversalOrder);
      traversalBuilder.SetStopEvaluator(stopEvaluator);
      traversalBuilder.SetReturnableEvaluator(returnableEvaluator);
      traversalBuilder.AddRelationshipCondition(new RelationshipCondition { Direction = direction, RelationshipName = relationshipName });
      var traverser = traversalBuilder.Build();
      return TraverseNodes(traverser);
    }

    public IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, string stopEvaluator, string returnableEvaluator, string relationshipName, Direction direction)
    {
      var traversalBuilder = new TraversalBuilder();
      traversalBuilder.SetOrder(traversalOrder);
      traversalBuilder.SetStopEvaluator(stopEvaluator);
      traversalBuilder.SetReturnableEvaluator(returnableEvaluator);
      traversalBuilder.AddRelationshipCondition(new RelationshipCondition { Direction = direction, RelationshipName = relationshipName });
      var traverser = traversalBuilder.Build();
      return TraverseNodes(traverser);
    }

    public IEnumerable<INode> TraverseNodes(TraversalOrder traversalOrder, StopEvaluator stopEvaluator, string returnableEvaluator, string relationshipName, Direction direction)
    {
      var traversalBuilder = new TraversalBuilder();
      traversalBuilder.SetOrder(traversalOrder);
      traversalBuilder.SetStopEvaluator(stopEvaluator);
      traversalBuilder.SetReturnableEvaluator(returnableEvaluator);
      traversalBuilder.AddRelationshipCondition(new RelationshipCondition { Direction = direction, RelationshipName = relationshipName });
      var traverser = traversalBuilder.Build();
      return TraverseNodes(traverser);
    }

    public long Id { get; private set; }

    public string Self { get; private set; }

    private string GetJsonData()
    {
      try
      {
        return Server.Get(Self + "/properties");
      }
      catch (WebException e)
      {
        throw new GraphDbException(string.Format("Failed to retrieve properties for node {0}", Id), e);
      }      
    }

    private static PropertyContainer CreatePropertyContainer(string jsonData)
    {
      var properties = new PropertiesParser().Parse(jsonData);
      return new PropertyContainer(properties);
    }

    private IPropertyContainer PropertyContainer 
    {
      get
      {
        if (_propertyContainer == null)
        {
          var jsonData = GetJsonData();
          _propertyContainer = CreatePropertyContainer(jsonData);
        }
        return _propertyContainer;
      }
    }

    public T GetProperty<T>(string key) 
    {
      return PropertyContainer.GetProperty<T>(key);
    }

    public T GetProperty<T>(string key, T defaultValue)
    {
      return PropertyContainer.GetProperty(key, defaultValue);
    }

    public IEnumerable<string> GetPropertyKeys()
    {
      return PropertyContainer.GetPropertyKeys();
    }

    public bool HasProperty(string key)
    {
      return PropertyContainer.HasProperty(key);
    }

    public object RemoveProperty(string key)
    {
      // Old value might have to be read before deleting it
      var oldValue = PropertyContainer.RemoveProperty(key);
      Server.Delete(Self + "/properties/" + key);
      return oldValue;
    }

    public void SetProperty<T>(string key, T value)
    {
      var encodedValue = new PropertyEncoder().Encode(value);
      Server.Put(Self + "/properties/" + key, encodedValue);
      PropertyContainer.SetProperty(key, value);
    }
  }
}