﻿using System;
using System.Collections.Generic;
using System.Linq;
using Neo4RestNet.Core;

namespace Neo4RestNet.Rest.Traverser
{  
  internal class TraversalBuilder
  {
    private ReturnableEvaluatorBuilder _returnableEvaluatorBuilder;
    private readonly IList<RelationshipConditionBuilder> _relationshipConditionBuilders = new List<RelationshipConditionBuilder>();
    private TraversalOrderBuilder _traversalOrderBuilder;
    private StopEvaluatorBuilder _stopEvaluatorBuilder;
    private MaxDepthBuilder _maxDepthBuilder;

    public void SetOrder(TraversalOrder traversalOrder)
    {
      _traversalOrderBuilder = new TraversalOrderBuilder(traversalOrder);
    }

    public void SetStopEvaluator(string javascript)
    {
      _stopEvaluatorBuilder = new StopEvaluatorBuilder(javascript);
    }

    public void SetStopEvaluator(StopEvaluator stopEvaluator)
    {
      switch (stopEvaluator)
      {
        case StopEvaluator.DepthOne:
          _maxDepthBuilder = new MaxDepthBuilder(1);
          break;
        case StopEvaluator.EndOfGraph:
          _maxDepthBuilder = new MaxDepthBuilder(Int32.MaxValue);
          break;
        default:
          throw new ArgumentOutOfRangeException("stopEvaluator");
      }
    }

    public void SetReturnableEvaluator(ReturnableEvaluator returnableEvaluator)
    {
      switch (returnableEvaluator)
      {
        case ReturnableEvaluator.All:
          _returnableEvaluatorBuilder = null;
          break;
        default:
          _returnableEvaluatorBuilder = new ReturnableEvaluatorBuilder(returnableEvaluator);
          break;          
      }
    }

    public void SetReturnableEvaluator(string javascript)
    {
      _returnableEvaluatorBuilder = new ReturnableEvaluatorBuilder(javascript);
    }

    public void ClearRelationshipConditions()
    {
      _relationshipConditionBuilders.Clear();
    }

    public void AddRelationshipCondition(RelationshipCondition condition)
    {
      _relationshipConditionBuilders.Add(new RelationshipConditionBuilder(condition));
    }

    public void AddRelationshipConditions(IEnumerable<RelationshipCondition> conditions)
    {
      _relationshipConditionBuilders.Concat(conditions.Select(c => new RelationshipConditionBuilder(c)));
    }

    public string Build()
    {
      var parts = new List<string>();

      if (_traversalOrderBuilder != null) parts.Add(_traversalOrderBuilder.Build());
      if (_relationshipConditionBuilders != null && _relationshipConditionBuilders.Count() > 0)
      {
        parts.Add(new RelationshipListConditionBuilder(_relationshipConditionBuilders).Build());
      }
      if (_stopEvaluatorBuilder != null) parts.Add(_stopEvaluatorBuilder.Build());
      if (_returnableEvaluatorBuilder != null) parts.Add(_returnableEvaluatorBuilder.Build());
      if (_maxDepthBuilder != null) parts.Add(_maxDepthBuilder.Build());

      return "{" + string.Join(",", parts.ToArray()) + "}";
    }
  }
}
