﻿using System;
using Neo4RestNet.ExpressionParser;
using NUnit.Framework;

namespace Neo4RestNet.Tests
{
  [TestFixture]
  public class ExpressionToJsTests
  {
    private ExpressionParser.ExpressionParser _expressionParser;

    [SetUp]
    public void Setup()
    {
      _expressionParser = new ExpressionParser.ExpressionParser(new AtomConverter(), new ExpressionEvaluator(new AtomConverter()));
    }


    // ReSharper disable InconsistentNaming
    [Test]
    public void true_constant()
    {
      var result = _expressionParser.ConvertToJs(r => true);
      Assert.That(result, Is.EqualTo("true"));
    }

    [Test]
    public void parameter_member_eq_to_int_constant()
    {
      var result = _expressionParser.ConvertToJs(r => r.Depth == 4);
      Assert.That(result, Is.EqualTo("position.depth() == 4"));      
    }

    [Test]
    public void parameter_member_gt_int_constant()
    {
      var result = _expressionParser.ConvertToJs(r => r.Depth > 1);
      Assert.That(result, Is.EqualTo("position.depth() > 1"));
    }

    [Test]
    public void parameter_member_call_eq_string_constant()
    {
      var result = _expressionParser.ConvertToJs(r => r.Node.GetProperty<string>("property") == "o");
      Assert.That(result, Is.EqualTo("position.node().getProperty('property') == 'o'"));      
    }

    [Test]
    public void condition_with_and()
    {
      var result = _expressionParser.ConvertToJs(r => r.Node.GetProperty<string>("property") == "o" && r.Node.GetProperty<int>("x") < 10);
      Assert.That(result, Is.EqualTo("position.node().getProperty('property') == 'o' && position.node().getProperty('x') < 10"));            
    }

    [Test]
    public void priod_add_with_comparison()
    {
      var result = _expressionParser.ConvertToJs(r => (r.Node.GetProperty<int>("property") + 1) >= (r.Node.GetProperty<int>("x") - 10));
      Assert.That(result, Is.EqualTo("position.node().getProperty('property') + 1 >= position.node().getProperty('x') - 10"));                  
    }

    [Test]
    public void several_conditions()
    {
      var result = _expressionParser.ConvertToJs(r => r.Depth > 4 && (r.Node.GetProperty<int>("x") != 0 || r.Depth == 1));
      Assert.That(result, Is.EqualTo("position.depth() > 4 && position.node().getProperty('x') != 0 || position.depth() == 1"));
    }

    [Test]
    public void reference_to_outer_constant()
    {
      const string outerVar = "x";
      var result = _expressionParser.ConvertToJs(r => r.Node.GetProperty<string>("x") == outerVar);
      Assert.That(result, Is.EqualTo("position.node().getProperty('x') == 'x'"));
    }

    [Test]
    public void calling_string_function_on_parameter()
    {
      var result = _expressionParser.ConvertToJs(r => r.Node.GetProperty<string>("x").ToLower() == "x");
      Assert.That(result, Is.EqualTo("position.node().getProperty('x').toLowerCase() == 'x'"));
    }

    [Test]
    public void calling_string_function_on_outer()
    {
      var outerVar = Guid.NewGuid().ToString();
      var result = _expressionParser.ConvertToJs(r => r.Node.GetProperty<string>("x") == outerVar.ToUpper());
      Assert.That(result, Is.EqualTo("position.node().getProperty('x') == '" + outerVar + "'.toUpperCase()"));
    }

    public static int GetMaxDepth()
    {
      return 4;
    }

    [Test]
    public void calling_static_function_with_no_parameters()
    {
      var result = _expressionParser.ConvertToJs(r => r.Depth <= GetMaxDepth());
      Assert.That(result, Is.EqualTo("position.depth() <= 4"));
    }

    // ReSharper restore InconsistentNaming
  }
}
