﻿using System;
using System.Linq;
using Neo4RestNet.Core;
using Neo4RestNet.Rest;
using NUnit.Framework;

namespace Neo4RestNet.Tests
{
  [TestFixture]
  public class RestIndexServiceTests
  {
    private IGraphDatabaseService _graphDb;
    private RestIndexService _indexService;

    [SetUp]
    public void Setup()
    {
      _graphDb = new RestGraphDatabaseService(Constants.DatabaseUrl);
      _indexService = new RestIndexService(Constants.DatabaseUrl);
    }

    // ReSharper disable InconsistentNaming
    [Test]
    public void add_node_to_index_should_succeed()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("indexedProp", node.Id);
      _indexService.AddIndex(node, "indexedProp", node.GetProperty<long>("indexedProp") );
      var nodes = _indexService.GetNodes("indexedProp", node.GetProperty<long>("indexedProp"));

      Assert.That(nodes.Where(n => n.Id == node.Id).Count(), Is.EqualTo(1));
    }

    [Test]
    public void remove_node_from_index_should_succeed()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("indexedProp", node.Id);
      _indexService.AddIndex(node, "indexedProp", node.GetProperty<long>("indexedProp"));
      _indexService.RemoveIndex(node, "indexedProp", node.GetProperty<long>("indexedProp"));
      var nodes = _indexService.GetNodes("indexedProp", node.GetProperty<long>("indexedProp"));

      Assert.That(nodes.Where(n => n.Id == node.Id).Count(), Is.EqualTo(0));
    }    

    [Test]
    public void add_multiple_nodes_to_index_should_succed()
    {
      var uniqueIndex = Guid.NewGuid().ToString();
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("multiindexedProp", uniqueIndex);
      _indexService.AddIndex(node, "multiindexedProp", uniqueIndex);
      node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("multiindexedProp", uniqueIndex);
      _indexService.AddIndex(node, "multiindexedProp", uniqueIndex);
      node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("multiindexedProp", uniqueIndex);
      _indexService.AddIndex(node, "multiindexedProp", uniqueIndex);
      var nodes = _indexService.GetNodes("multiindexedProp", uniqueIndex);

      Assert.That(nodes.Count(), Is.EqualTo(3));      
    }

    [Test]
    public void get_single_node_should_succeed()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("indexedProp", node.Id);
      _indexService.AddIndex(node, "indexedProp", node.GetProperty<long>("indexedProp"));
      var nodeFromIndex = _indexService.GetSingleNode("indexedProp", node.GetProperty<long>("indexedProp"));

      Assert.NotNull(nodeFromIndex);
      Assert.That(nodeFromIndex.Id, Is.EqualTo(node.Id));
    }

    [Test]
    public void get_single_node_with_no_match_should_return_null()
    {
      var nodeFromIndex = _indexService.GetSingleNode("my faked key", 1);
      Assert.IsNull(nodeFromIndex);
    }

    [Test]
    public void get_single_node_with_more_than_one_hit_should_throw()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      _indexService.AddIndex(node, "xxx", "xxx");
      node = _graphDb.CreateNode(Constants.TestNode);
      _indexService.AddIndex(node, "xxx", "xxx");

      Assert.Throws<GraphDbException>(() => _indexService.GetSingleNode("xxx", "xxx"));
    }

    // ReSharper restore InconsistentNaming
  }
}
