﻿using System.Linq;
using Neo4RestNet.Core;
using Neo4RestNet.Rest;
using NUnit.Framework;

namespace Neo4RestNet.Tests
{
  // Todo: Keep track of all nodes created and clean up after us when the tests are done

  [TestFixture]
  public class RestNodeTests
  {
    private IGraphDatabaseService _graphDb;

    [SetUp]
    public void Setup()
    {
      _graphDb = new RestGraphDatabaseService(Constants.DatabaseUrl);
    }

    // ReSharper disable InconsistentNaming
    [Test]
    public void delete_node_should_succeed()
    {
      // Create a node to be deleted
      var node = _graphDb.CreateNode(Constants.TestNode);
      // Read it back again to be sure that it really exists
      node = _graphDb.GetNode(node.Id);
      node.Delete();
      var id = node.Id;

      // We assume here that we will get an exception when reading a non-existing node
      Assert.Throws<GraphDbException>(() => _graphDb.GetNode(id).HasProperty("test"));
    }

    [Test]
    public void delete_non_existing_node_should_throw()
    {
      // Create a node to be deleted
      var node = _graphDb.CreateNode(Constants.TestNode);
      // and delete it
      node.Delete();

      // Delete it again and we will surely get an exception
      Assert.Throws<GraphDbException>(node.Delete);
    }

    [Test]
    public void create_relation_should_succeed()
    {
      var fromNode = _graphDb.CreateNode(Constants.TestNode);
      var toNode = _graphDb.CreateNode(Constants.TestNode);

      fromNode.CreateRelationshipTo(toNode, "test");

      var relatedNodes = fromNode.GetRelationships(Direction.Outgoing);

      Assert.That(relatedNodes.Count(), Is.EqualTo(1));
      Assert.That(relatedNodes.First().StartNode.Id, Is.EqualTo(fromNode.Id));
      Assert.That(relatedNodes.First().EndNode.Id, Is.EqualTo(toNode.Id));
    }

    [Test]
    public void create_two_relations_should_succeed()
    {
      var fromNode = _graphDb.CreateNode(Constants.TestNode);
      var toNode1 = _graphDb.CreateNode(Constants.TestNode);
      var toNode2 = _graphDb.CreateNode(Constants.TestNode);

      fromNode.CreateRelationshipTo(toNode1, "test");
      fromNode.CreateRelationshipTo(toNode2, "test");

      var relatedNodes = fromNode.GetRelationships(Direction.Outgoing);

      Assert.That(relatedNodes.Count(), Is.EqualTo(2));
      var numberOfNames = relatedNodes.Select(r => r.Name).Distinct().Count();
      Assert.That(numberOfNames, Is.EqualTo(1));
      Assert.That(relatedNodes.First().StartNode.Id, Is.EqualTo(fromNode.Id));
      Assert.That(relatedNodes.First().EndNode.Id, Is.EqualTo(toNode1.Id));
      Assert.That(relatedNodes.Last().StartNode.Id, Is.EqualTo(fromNode.Id));
      Assert.That(relatedNodes.Last().EndNode.Id, Is.EqualTo(toNode2.Id));
    }

    [Test]
    public void get_all_relationships_should_succeed()
    {
      var fromNode = _graphDb.CreateNode(Constants.TestNode);
      var toNode = _graphDb.CreateNode(Constants.TestNode);

      // fromNode will have 1 outgoing and 1 ingoing relation with different names
      fromNode.CreateRelationshipTo(toNode, "out");
      toNode.CreateRelationshipTo(fromNode, "in");
      var relationships = fromNode.GetRelationships();

      Assert.That(relationships.Count(), Is.EqualTo(2));
    }

    [Test]
    public void get_single_relationship_when_no_relations_should_return_null()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      var relationship = node.GetSingleRelationship("x", Direction.All);

      Assert.That(relationship, Is.Null);
    }

    [Test]
    public void get_single_relationship_should_succeed()
    {
      var fromNode = _graphDb.CreateNode(Constants.TestNode);
      var toNode = _graphDb.CreateNode(Constants.TestNode);

      // fromNode will have 1 outgoing and 1 ingoing relation with different names
      fromNode.CreateRelationshipTo(toNode, "out");

      var relationship = fromNode.GetSingleRelationship("out", Direction.Outgoing);      

      Assert.That(relationship, Is.Not.Null);
    }

    [Test]
    public void get_single_relationship_with_more_than_one_should_throw()
    {
      var fromNode = _graphDb.CreateNode(Constants.TestNode);
      var toNode = _graphDb.CreateNode(Constants.TestNode);

      fromNode.CreateRelationshipTo(toNode, "x");
      toNode.CreateRelationshipTo(fromNode, "x");

      Assert.Throws<GraphDbException>(() => fromNode.GetSingleRelationship("x", Direction.All));
    }

    [Test]
    public void get_relationships_by_multiple_names_should_succeed()
    {
      var fromNode = _graphDb.CreateNode(Constants.TestNode);
      var toNode = _graphDb.CreateNode(Constants.TestNode);

      fromNode.CreateRelationshipTo(toNode, "out");
      toNode.CreateRelationshipTo(fromNode, "in");
      var relationships = fromNode.GetRelationships(new System.Collections.Generic.List<string> { "out", "in"});

      Assert.That(relationships.Count(), Is.EqualTo(2));
    }

    [Test]
    public void get_relationsships_by_multiple_names_and_exclude_one_should_succeed()
    {
      var startnode = _graphDb.CreateNode(Constants.TestNode);
      var buddyNode = _graphDb.CreateNode(Constants.TestNode);
      var familyNode = _graphDb.CreateNode(Constants.TestNode);
      var enemyNode = _graphDb.CreateNode(Constants.TestNode);
      startnode.CreateRelationshipTo(buddyNode, "buddy");
      startnode.CreateRelationshipTo(familyNode, "family");
      startnode.CreateRelationshipTo(enemyNode, "enemy");
      
      // Traverse buddy & family relations
      var relationships = startnode.GetRelationships(new System.Collections.Generic.List<string> { "buddy", "family" });

      Assert.That(relationships.Count(), Is.EqualTo(2));
      var endNodes = relationships.Select(r => r.EndNode);
      Assert.That(endNodes.Where(n => n.Id == familyNode.Id).Count(), Is.EqualTo(1));
      Assert.That(endNodes.Where(n => n.Id == buddyNode.Id).Count(), Is.EqualTo(1));
    }

    [Test]
    public void get_property_on_node_should_succeed()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      var valBeforeGet = node.GetProperty<string>("test");
      node = _graphDb.GetNode(node.Id);
      var valAfterGet = node.GetProperty<string>("test");

      Assert.That(valBeforeGet, Is.EqualTo(valAfterGet));
      Assert.That(valBeforeGet, Is.EqualTo("true"));
    }

    [Test]
    public void set_property_on_node_should_be_saved()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("newProperty", 0);
      node = _graphDb.GetNode(node.Id);

      Assert.That(node.GetProperty<int>("newProperty"), Is.EqualTo(0));      
    }

    [Test]
    public void set_double_should_succeed()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty<double>("newProperty", 1.5);
      node = _graphDb.GetNode(node.Id);
      var prop = node.GetProperty<double>("newProperty");

      Assert.That(prop, Is.EqualTo(1.5));
    }


    // Problem with arrays on rest server?
    //[Test]
    //public void set_array_of_int_should_succeed()
    //{
    //  var node = _graphDb.CreateNode(Constants.TestNode);
    //  node.SetProperty("newProperty", new []{1,2,3});
    //  node = _graphDb.GetNode(node.Id);
    //  var prop = node.GetProperty<int[]>("newProperty");

    //  Assert.That(prop[0], Is.EqualTo(1));
    //  Assert.That(prop[1], Is.EqualTo(2));
    //  Assert.That(prop[2], Is.EqualTo(3));
    //}

    [Test]
    public void remove_property_on_node_should_succeed()
    {
      var node = _graphDb.CreateNode(Constants.TestNode);
      node.SetProperty("propertyToDelete", 0);
      var oldValue = node.RemoveProperty("propertyToDelete");
      node = _graphDb.GetNode(node.Id);

      Assert.That(oldValue, Is.EqualTo(0));
      Assert.That(node.HasProperty("propertyToDelete"), Is.False);      
    }

    [Test]
    public void add_node_with_propertycontainer_should_succeed()
    {
      var properties = new PropertyContainer();
      properties.SetProperty("HeyHo", "Lets go");
      properties.SetProperty("Count", 1234);

      var node = _graphDb.CreateNode(properties);

      Assert.That(node.GetProperty<string>("HeyHo"), Is.EqualTo("Lets go"));
      Assert.That(node.GetProperty<int>("Count"), Is.EqualTo(1234));
    }

    // ReSharper restore InconsistentNaming
  }
}
