﻿using System.Linq;
using Neo4RestNet.Core;
using Neo4RestNet.Rest;
using NUnit.Framework;

namespace Neo4RestNet.Tests
{
  [TestFixture]
  public class RestNodeTraversalTests
  {
    private IGraphDatabaseService _graphDb;
    private const string SimpleTraverser = @"{""order"":""breadth first""}";

    [SetUp]
    public void Setup()
    {
      _graphDb = new RestGraphDatabaseService(Constants.DatabaseUrl);
    }

    // ReSharper disable InconsistentNaming
    [Test]
    public void simple_traverse_should_succeed()
    {
      var startnode = _graphDb.CreateNode(Constants.TestNode);
      var visited1 = _graphDb.CreateNode(Constants.TestNode);
      var visited2 = _graphDb.CreateNode(Constants.TestNode);

      startnode.CreateRelationshipTo(visited1, "test");
      startnode.CreateRelationshipTo(visited2, "test");

      var visitedNodes = startnode.TraverseNodes(SimpleTraverser).ToList();

      Assert.That(visitedNodes.Count(), Is.EqualTo(2));
      Assert.IsTrue(visitedNodes.Where(n => n.Id == visited1.Id).Count() == 1);
      Assert.IsTrue(visitedNodes.Where(n => n.Id == visited2.Id).Count() == 1);
    }

    [Test]
    public void traverse_with_depth_one()
    {
      var startnode = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthOne = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthTwo = _graphDb.CreateNode(Constants.TestNode);

      startnode.CreateRelationshipTo(nodeAtDepthOne, "test");
      nodeAtDepthOne.CreateRelationshipTo(nodeAtDepthTwo, "test");

      var nodes = startnode.TraverseNodes(TraversalOrder.DepthFirst, StopEvaluator.DepthOne, ReturnableEvaluator.All, "test",
                              Direction.All);

      Assert.That(nodes.Count(), Is.EqualTo(1));
      Assert.That(nodes.First().Id, Is.EqualTo(nodeAtDepthOne.Id));
    }

    [Test]
    public void traverse_with_end_of_graph()
    {
      var startnode = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthOne = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthTwo = _graphDb.CreateNode(Constants.TestNode);
      startnode.CreateRelationshipTo(nodeAtDepthOne, "test");
      nodeAtDepthOne.CreateRelationshipTo(nodeAtDepthTwo, "test");

      var nodes = startnode.TraverseNodes(TraversalOrder.DepthFirst, StopEvaluator.EndOfGraph, ReturnableEvaluator.All, "test",
                              Direction.All);

      Assert.That(nodes.Count(), Is.EqualTo(2));
      Assert.That(nodes.Where(n => n.Id == nodeAtDepthOne.Id).Count(), Is.EqualTo(1));
      Assert.That(nodes.Where(n => n.Id == nodeAtDepthTwo.Id).Count(), Is.EqualTo(1));
    }

    [Test]
    public void traverse_with_js_as_stop_evaluator()
    {
      var startnode = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthOne = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthTwo = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthThree = _graphDb.CreateNode(Constants.TestNode);
      startnode.CreateRelationshipTo(nodeAtDepthOne, "test");
      nodeAtDepthOne.CreateRelationshipTo(nodeAtDepthTwo, "test");
      nodeAtDepthTwo.CreateRelationshipTo(nodeAtDepthThree, "test");
      nodeAtDepthTwo.SetProperty("stop", true);
      var nodes = startnode.TraverseNodes(TraversalOrder.DepthFirst, "position.node().hasProperty('stop') && position.node().getProperty('stop')==true;",
                                          ReturnableEvaluator.All, "test", Direction.All);
      Assert.That(nodes.Count(), Is.EqualTo(2));
      Assert.That(nodes.Where(n => n.Id == nodeAtDepthOne.Id).Count(), Is.EqualTo(1));
      Assert.That(nodes.Where(n => n.Id == nodeAtDepthTwo.Id).Count(), Is.EqualTo(1));
    }

    [Test]
    public void traverse_with_js_as_returnable_evaluator()
    {
      var startnode = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthOne = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthTwo = _graphDb.CreateNode(Constants.TestNode);
      var nodeAtDepthThree = _graphDb.CreateNode(Constants.TestNode);
      startnode.CreateRelationshipTo(nodeAtDepthOne, "test");
      nodeAtDepthOne.CreateRelationshipTo(nodeAtDepthTwo, "test");
      nodeAtDepthTwo.CreateRelationshipTo(nodeAtDepthThree, "test");
      startnode.SetProperty("pickme", 1);
      nodeAtDepthThree.SetProperty("pickme", 1);

      var nodes = startnode.TraverseNodes(TraversalOrder.DepthFirst, StopEvaluator.EndOfGraph,
                                          "position.node().hasProperty('pickme')", "test", Direction.All);
      Assert.That(nodes.Count(), Is.EqualTo(2));
      Assert.That(nodes.Where(n => n.Id == startnode.Id).Count(), Is.EqualTo(1));
      Assert.That(nodes.Where(n => n.Id == nodeAtDepthThree.Id).Count(), Is.EqualTo(1));
    }

    // ReSharper restore InconsistentNaming
  }
}
