﻿using System.Collections.Generic;
using Neo4RestNet.Core;
using Neo4RestNet.Rest.Traverser;
using NUnit.Framework;

namespace Neo4RestNet.Tests.Traverser
{
  [TestFixture]
  public class TraverserBuilderTests
  {
    private static string TrimResult(string buildResult)
    {
      // Trim for easier comparison
      buildResult = buildResult.Trim();
      buildResult = buildResult.Replace("  ", string.Empty);
      buildResult = buildResult.Replace("\n", string.Empty);
      buildResult = buildResult.Replace("\t", string.Empty);
      buildResult = buildResult.Replace("\r", string.Empty);
      return buildResult;
    }

    // ReSharper disable InconsistentNaming
    [Test]
    public void creating_condition_for_relationship_direction_all_should_succeed()
    {
      var builder =
        new RelationshipConditionBuilder(new RelationshipCondition {Direction = Direction.All, RelationshipName = "x"});
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"{""type"": ""x"",""direction"": ""all""}"));
    }

    [Test]
    public void creating_condition_for_relationship_direction_in_should_succeed()
    {
      var builder =
        new RelationshipConditionBuilder(new RelationshipCondition { Direction = Direction.Incoming, RelationshipName = "x" });
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"{""type"": ""x"",""direction"": ""in""}"));
    }

    [Test]
    public void creating_condition_for_relationship_direction_out_should_succeed()
    {
      var builder =
        new RelationshipConditionBuilder(new RelationshipCondition { Direction = Direction.Outgoing, RelationshipName = "x" });
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"{""type"": ""x"",""direction"": ""out""}"));
    }

    [Test]
    public void creating_condition_for_max_depth_should_succeed()
    {
      var builder = new MaxDepthBuilder(1);
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""max depth"": 1"));    
    }

    [Test]
    public void creating_list_of_relationships_conditions_should_succeed()
    {
      var relationshipCondition =
        new RelationshipConditionBuilder(new RelationshipCondition {Direction = Direction.All, RelationshipName = "x"});
      var builder = new RelationshipListConditionBuilder(new List<RelationshipConditionBuilder> {
                                                                                                  relationshipCondition,
                                                                                                  relationshipCondition
                                                                                                });
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""relationships"": [{""type"": ""x"",""direction"": ""all""},{""type"": ""x"",""direction"": ""all""}]"));
    }

    [Test]
    public void creating_builtin_returnableevaluator_should_succeed()
    {
      var builder = new ReturnableEvaluatorBuilder(ReturnableEvaluator.All);
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""return filter"": {""language"": ""builtin"",""name"": ""all""}"));          
    }

    [Test]
    public void creating_js_returnableevaluator_should_succeed()
    {
      var builder = new ReturnableEvaluatorBuilder("function(){}");
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""return filter"": {""language"": ""javascript"",""body"": ""function(){}""}"));
    }

    [Test]
    public void creating_stopevaluator_should_succeed()
    {
      var builder = new StopEvaluatorBuilder("function(){}");
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""prune evaluator"": {""language"": ""javascript"",""body"": ""function(){}""}"));      
    }

    [Test]
    public void creating_depth_first_order_should_succeed()
    {
      var builder = new TraversalOrderBuilder(TraversalOrder.DepthFirst);
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""order"": ""depth first"""));
    }

    [Test]
    public void creating_breadth_first_order_should_succeed()
    {
      var builder = new TraversalOrderBuilder(TraversalOrder.BreadthFirst);
      var buildResult = builder.Build();
      buildResult = TrimResult(buildResult);

      Assert.That(buildResult, Is.EqualTo(@"""order"": ""breadth first"""));
    }

    // ReSharper restore InconsistentNaming
  }
}
